#include "Common/StdAfx.h"
#include "Compress.h"
#include "Kernel.h"
#include "../Strings.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*
 * $Source: /usr/cvsroot/cdx/src/cdx/CPacker.cpp,v $
 * $Author: Mindcry $
 *
 * $Log: CPacker.cpp,v $
 * Revision 1.4  2000/03/10 19:44:00  Avalon
 * Added function PackRamAppend
 * Added support for compressed data files
 * Reworked decode routine for extra speed
 *
 * Revision 1.1  1999/07/17 11:56:00  Mindcry
 * Initial Revision to CVS.
 * for expand to Surface you need a special compressor!
 */

CPacker pack;

CPacker::CPacker (void) {
    int i, c;
    sym_cum[N_CHAR] = 0;
    for (i = N_CHAR; i >= 1; i--) {
        c = i - 1;
        char_to_sym [c]   = i;
        sym_to_char [i]   = c;
        sym_freq    [i]   = 1;
        sym_cum     [i-1] = sym_cum[i] + sym_freq[i];
    }
    sym_freq     [0] = 0;
    position_cum [N] = 0;

    for (i = N; i >= 1; i--)
        position_cum[i - 1] = position_cum[i] + 10000 / (i + 200);
    memcpy (stdtable, char_to_sym, (N_CHAR * 4) + N + 4);
}

#ifdef _DEBUG

void CPacker::ReadStr (FILE *file, CString* string) {
    unsigned char c = 0;
    CString       s1;
    while (c < 32) c = fgetc (file);
    while (c > 31 && ((feof(file) | ferror(file)) == 0)) {
        s1 += c;
        c = fgetc (file);
    }
    string->operator = (s1);
}

int CPacker::ReadNum (FILE *file) {
    int c = 0,
        n = 0;
    while ((c < '0' || c > '9') && c >= 0) c = fgetc (file);
    while (c >='0' && c <='9') {
        n = (n * 10) + (c - '0');
        c = fgetc (file);
    }
    if (ferror (file)) return 0;
    return n;
}

void CPacker::MakeMediaLibrary (const char *thedir) {
    FILE    *list,
            *file,
            *MediaLib;
    CString s1;
    int     ver = 0,
            i   = 0,
            j;
    list = fopen ("MediaLib.h", "ra");
    if (list) {
        while (s1.Find ("LIBVERSION") == -1 && !feof(list)) ReadStr (list, &s1);
        for (i=0; i < s1.GetLength (); i++) {
            j = s1.GetAt (i);
            if (j >= '0' && j <= '9') {
                ver *= 10; ver += (j - '0');
            } else if (ver) break;
        }
        ver++;
    }

    s1       = DataDir;
    s1      += AppID;
    s1      += ".dat";
    NumRefs  = 1;
    NumValid = 1;
    MediaLib = fopen (s1,           "wb");
    list     = fopen ("MediaLib.h", "wa");
    s1       = AppID; s1 += ".h";
    file     = fopen (s1,           "rb");
    fprintf (list, "/" "* Do not edit the file - it will be overwritten */\n\n");
    fprintf (list, "#" "ifdef MEDIALIB_H\n#error Double include\n#" "endif\n");
    fprintf (list, "#" "define MEDIALIB_H\n\n#" "define LIBVERSION %d\nsigned long MediaRef [] = {\n", ver);

    fprintf (MediaLib, AppName);
    fwrite  (&ver, sizeof (ver), 1, MediaLib);
    s1 = "Strings";
    fprintf (list, "\n/" "*------ String tables ------*/\n");
    InsertInLibrary (MediaLib, list, &s1, RES_STR);
    if (thedir) theApp.ChangeDirectory (thedir);
    fprintf (list, "\n/" "*------ Bitmap images ------*/\n");
    while (s1.Find ("BitmapFiles") == -1) ReadStr (file, &s1);
    while (1) {
        ReadStr (file, &s1); if (s1.Find ("MAXBMP") != -1) break;
        s1 = s1.Mid (s1.Find ('_') + 1); s1.Remove  (',');
        InsertInLibrary (MediaLib, list, &s1, RES_BMP);
    }
    #ifndef NO_ANIM
    fprintf (list, "\n/" "*------ Sprite images ------*/\n");
    while (s1.Find ("SpriteFiles") == -1) ReadStr (file, &s1);
    for (i=Str_AnimDir + 1; i<Str_AniFirst; i++) {
        s1 = StrTable[i];
        InsertInLibrary (MediaLib, list, &s1, RES_ANIM);
    }
    #endif
    fprintf (list, "\n/" "*------- JPEG images -------*/\n");
    while (s1.Find ("JpegFiles") == -1) ReadStr (file, &s1);
    while (1) {
        ReadStr (file, &s1); if (s1.Find ("MAXJPEG") != -1) break;
        s1 = s1.Mid (s1.Find ('_') + 1); s1.Remove  (',');
        InsertInLibrary (MediaLib, list, &s1, RES_JPEG);
    }
    fprintf (list, "\n/" "*------- Midi musics -------*/\n");
    while (s1.Find ("MidiFiles") == -1) ReadStr (file, &s1);
    while (1) {
        ReadStr (file, &s1); if (s1.Find ("MAXMIDI") != -1) break;
        s1 = s1.Mid (s1.Find ('_') + 1); s1.Remove  (',');
        InsertInLibrary (MediaLib, list, &s1, RES_MIDI);
    }
    fprintf (list, "\n/" "*------- Wave sounds -------*/\n");
    while (s1.Find ("WaveFiles") == -1) ReadStr (file, &s1);
    while (1) {
        ReadStr (file, &s1);
        if (s1.Find ("MAXWAVE") != -1) break;
        if (s1.Find ("#" "define") != -1) continue;
        if (s1.Find ('(') != -1) {
            s1 = s1.Mid (s1.Find ('(') + 1);
            s1.Remove (',');
            for (i=1; i<8; i++) {
                s1.SetAt (s1.GetLength () - 1, '0' + i);
                InsertInLibrary (MediaLib, list, &s1, RES_WAVE);
            }
        } else {
            s1 = s1.Mid (s1.Find ('_') + 1);
            s1.Remove  (',');
            InsertInLibrary (MediaLib, list, &s1, RES_WAVE);
        }
    }
    fprintf (list, "\n/" "*-------- Helpfiles --------*/\n");
    while (s1.Find ("HelpFiles") == -1) ReadStr (file, &s1);
    while (1) {
        ReadStr (file, &s1); if (s1.Find ("MAXHELP") != -1) break;
        s1 = s1.Mid (s1.Find ('_') + 1); s1.Remove  (',');
        InsertInLibrary (MediaLib, list, &s1, RES_FILE);
    }
    fprintf (list, "\n/" "*-------- Miscfiles --------*/\n");
    for (i=0; i<=Str_MiscLast - Str_MiscFile; i++) {
        s1 = StrTable[Str_MiscFile + i];
        InsertInLibrary (MediaLib, list, &s1, RES_FILE);
    }
    theApp.HandleMakeLib (list, MediaLib);
    fprintf (list, "\n/" "*----- Size of library -----*/\n%10d, -1 };\n", ftell(MediaLib));
    fprintf (list, "/" "*%8d  resources,\n *%8d  references\n", NumValid, NumRefs);
    fprintf (list, " *%8d%% done\n */\n", (NumValid*100) / NumRefs);
    if (ferror (file))     TRACE0 ("MakeMediaLib: source file error.\n");
    if (ferror (list))     TRACE0 ("MakeMediaLib: list file error.\n");
    if (ferror (MediaLib)) TRACE0 ("MakeMediaLib: library file error.\n");
    fclose (list);
    fclose (file);
    fclose (MediaLib);
    AfxMessageBox ("Created media library - please recompile.");
}

void CPacker::InsertInLibrary (FILE *lib, FILE *list, CString *name, UCHAR restype) {
    char         format [] = "%10d, /x %s\n"; format[7] = '/';
    static char  cnt = 0;
    int          i;
    char         tag[4];
    bool         docompress = false;
    ULONG        size,
                 offs;
    FILE         *file;
    char         c;
    CString      s1;
    cnt = (cnt + 1) % 7;
    if (name->Find ("EXTRA") > -1) {
        name->Delete (name->GetLength () - 6, 6);
    }
    if (name->Find ("/" "/=") != -1) {
        NumRefs  ++;
        NumValid ++;
        s1 = *name;
        s1.Delete (0, s1.Find ("/" "/=") + 4);
        s1.Delete (s1.Find ('.'), s1.GetLength () - s1.Find ('.'));
        i = MkNum (s1) | ALTLIB;
        i ^= cnt * GARBLE;
        s1 = *name;
        s1.Delete (s1.Find ("/" "/="), s1.GetLength () - s1.Find ("/" "/="));
        s1.TrimRight (" "); 
        s1.Insert    (0, "*");
        TRACE1 ("- %s\n", s1);
        if (list) fprintf (list, format, i, s1);
        return;
    }
    if (name->Find ('/') != -1) {
        name->Delete (name->Find ('/'), name->GetLength () - name->Find ('/'));
        name->TrimRight (' ');
    }
    TRACE1 ("- %s\n", *name);
    switch (restype) {
        case RES_FILENC:
        case RES_FILE: if (name->Find (RESDIR) == -1) s1 = RESDIR "Misc\\";
                       s1 += *name;
                       break;
        case RES_FILET:s1 = RESDIR "World\\"; s1 += *name;               break;
        case RES_BMP:  s1 = RESDIR;           s1 += *name; s1 += ".bmp"; break;
        case RES_JPEG: s1 = RESDIR "Jpegs\\"; s1 += *name; s1 += ".jpg"; break;
        case RES_MIDI: s1 = RESDIR "Midis\\"; s1 += *name; s1 += ".mid"; break;
        case RES_ANIM: s1 = RESDIR "Anim\\";  s1 += *name; s1 += ".spr"; break;
        case RES_WAVE: s1 = RESDIR "Waves\\"; s1 += *name; 
                       if (s1.GetLength () > 18) s1.Delete (18, s1.GetLength () - 18);
                       s1 += ".wav"; break;
        case RES_STR:  if (list) fprintf (list, format, ftell(lib), "English");
                       InitBuffer ();
                       theApp.WriteStrTable ();
                       PackRamAppend (Buffer, textsize, lib);
                       NumRefs++;
                       NumValid++;
                       return;
    }
    if (restype != RES_FILET) s1.Replace (' ', '_');
    s1.Replace (".url", ".txt");
    NumRefs++;
    file = fopen (s1, "rb");
    if (file == NULL) {
        if (theApp.allresources) {
            s1.Insert (s1.ReverseFind ('\\') + 1, "FullVer\\");
            file = fopen (s1, "rb");
        }
        if (file == NULL) {
            s1.Insert (s1.ReverseFind ('\\') + 1, '_');
            if ((file = fopen (s1, "rb")) == NULL) {
                if (list) fprintf (list, format, 0, *name);
                return;
            }
        }
    }
    fseek (file, 0, SEEK_END); size = ftell (file);
    if (size > MAXLIBSIZE) {
        s1 =  "This file is too large: "; s1 += *name;
        s1 += "(size = "; s1 += MkStr (size);
        s1 += ", buffer = "; s1 += MkStr (MAXLIBSIZE);
        s1 += ")"; AfxMessageBox (s1);
        if (list) fprintf (list, format, 0, *name);
        return;
    }
    if (list) {
        i = ftell(lib);
        i ^= cnt * GARBLE;
        fprintf (list, format, i, *name);
    }
    NumValid++;
    switch (restype) {
        case RES_BMP:
            /* see BITMAPINFOHEADER and BITMAPFILEHEADER for details */
            fseek (file,  2, SEEK_SET);                          /* id 'BM'     */
            fread (&size, 4, 1, file);                           /* size        */
            fseek (file,  4, SEEK_CUR);                          /* reserved    */
            fread (&offs, 4, 1, file);                           /* offset      */
            fseek (file,  4, SEEK_CUR);                          /* size again  */
            fread (tag,   4, 1, file);  fwrite (tag, 4, 1, lib); /* width       */
            fread (tag,   4, 1, file);  fwrite (tag, 4, 1, lib); /* height      */
            fseek (file,  2, SEEK_CUR);                          /* planes      */
            fread (tag,   2, 1, file);                           /* bits/pixel  */
            if (tag[0] != 24) {
                s1 = "BMP file should have 24-bit colors: ";
                s1 += *name; AfxMessageBox (s1);
                if (list) fprintf (list, format, 0, *name);
                return;
            }
            fread (tag,   4, 1, file);                           /* compression */
            if (tag[0]) {
                s1 = "BMP file is compressed (possibly RLE): ";
                s1 += *name; AfxMessageBox (s1);
                if (list) fprintf (list, format, 0, *name);
                return;
            }
            docompress = true;
            fseek (file, offs, SEEK_SET); size -= offs; break;
        case RES_JPEG:
            fseek (file, 6, SEEK_SET); fread (tag, 4, 1, file);
            if (tag[0]=='J' && tag[1]=='F' && tag[2]=='I' && tag[3]=='F') {
                s1 = "JPEG file needs to have its headers stripped: ";
                s1 += *name; AfxMessageBox (s1);
                return;
            }
            fseek (file,  4, SEEK_SET);
            size -= 4;
            break;
        case RES_FILET:
            rewind (file);
            for (i=0; AppName[i]; i++, size--) {
                c = fgetc (file);
                if (c != AppName[i]) {
                    s1 = "Level file has wrong format: ";
                    s1 += *name;
                    TRACE1 ("%s\n", s1);
                    if (list) fprintf (list, format, 0, *name);
                    return;
                }
            }
            fgetc (file);
            size --;
            break;
        case RES_FILE:
        case RES_MIDI:
        case RES_WAVE:
        case RES_ANIM:
            docompress = true;
            rewind (file);
            break;
        case RES_FILENC:
            rewind (file);
            break;
    }
    fread                   (Buffer, 1, size, file);
    if (!docompress) fwrite (Buffer, 1, size, lib);
    else PackRamAppend      (Buffer,    size, lib);
    if (ferror (file)) TRACE0 ("InsertInLibrary: file error.\n");
    fclose (file);
}

/* Compresses the buffer and append the data to a file; only used in MediaLib */
unsigned CPacker::PackRamAppend (BYTE *in, DWORD sz, FILE *out) {
    unsigned i, checksum = 0;
    fout     = out;
    textsize = sz;
    ramptr   = in;
    if (in == NULL) ramptr = Buffer;
    CalcChecksum (checksum);
    fwrite (&checksum, sizeof (checksum), 1, fout);
    fwrite (&textsize, sizeof (textsize), 1, fout);
    if (ferror (fout)) return Str_ErrWrite;
    Encode ();
    return 0;
}

#else

void CPacker::MakeMediaLibrary (const char*) {}

#endif

/* Read data from a file into a new memory array, and decompress it */
char *CPacker::ExpandToRam (FILE *in) {
    unsigned i, checksum;
    fin = in;
    fread (&checksum, sizeof (checksum), 1, fin);
    fread (&textsize, sizeof (textsize), 1, fin);
    if (ferror (fin)) return NULL;
    ramptr = new BYTE [textsize + 3];
    if (ramptr == NULL) {
        theApp.runlevel = Run_MemError;
        return NULL;
    }
    Decode ();
    ramptr[textsize] = 0; ramptr[textsize + 1] = 0; ramptr[textsize + 2] = 0;
    CalcChecksum (checksum);
    if (checksum) {
        TRACE1 ("Checksum failure: %x\n", checksum);
        delete [] ramptr;
        return NULL;
    }
    return (char*) ramptr;
}

/* Initialize the buffer; next, use Write functions, then WriteBuffer */
void CPacker::InitBuffer (void) {
    ramptr   = Buffer;
    textsize = 0;
}

void CPacker::SeekTo (unsigned offset) {
    unsigned i;
    if (offset > MAXLIBSIZE) return;
    if (offset > textsize) {
        for (i=textsize; i<offset; i++) Buffer[i] = 0;
        textsize = offset;
    }
    ramptr = Buffer + offset;
}

/* Compress the data in the buffer and write it to a file */
unsigned CPacker::WriteBuffer (const char *fname) {
    unsigned i, checksum = 0;
    fout = fopen (fname, "wb");
    if (fout == NULL) return Str_ErrOpenW;
    for (i=0; AppName[i] != 0; i++) fputc (AppName[i], fout);
    fputc (0, fout);
    ramptr = Buffer;
    CalcChecksum (checksum);
    fwrite (&checksum, sizeof (checksum), 1, fout);
    fwrite (&textsize, sizeof (textsize), 1, fout);
    if (ferror (fout)) {
        fclose (fout);
        return Str_ErrWrite;
    }
    Encode ();
    fclose (fout);
    return 0;
}

/* Write data in the buffer to a file without compression */
unsigned CPacker::WriteBufferPure (const char *fname) {
    fout = fopen (fname, "wb");
    if (fout == NULL) return Str_ErrOpenW;
    fwrite (Buffer, textsize, 1, fout);
    if (ferror (fout)) {
        fclose (fout); return Str_ErrWrite;
    }
    fclose (fout);
    return 0;
}

/* Read from the medialib into the buffer and decompress the data */
unsigned CPacker::ReadBufCore (void) {
    unsigned i, checksum;
    fread (&checksum, sizeof (checksum), 1, fin);
    fread (&textsize, sizeof (textsize), 1, fin);
    if (ferror (fin) || textsize > MAXLIBSIZE) {
        textsize = 0;
        fclose (fin);
        return Str_ErrRead;
    }
    ramptr = Buffer;
    Decode ();
    CalcChecksum (checksum);
    ramptr = Buffer;
    if (checksum) {
        textsize = 0;
        TRACE1 ("Checksum failure: %x\n", checksum);
        return Str_Checksum;
    }
    return 0;
}

/* Read from a file into the buffer and decompress the data */
unsigned CPacker::ReadBuffer (const char *fname, const char *fileID) {
    unsigned i;
    char     c;
    if (fileID == NULL) fileID = AppName;
    if (fname == NULL) {
        fin = theApp.MediaLib;
        if (fin == NULL) return Str_ErrOpenR;
    } else {
        fin = fopen (fname, "rb");
        if (fin == NULL) return Str_ErrOpenR;
        for (i=0; fileID[i] != 0; i++) {
            c = fgetc (fin);
            if (c != fileID[i]) {
                fclose (fin); return Str_Checksum;
            }
        }
        fgetc (fin);
    }
    i = ReadBufCore ();
    if (fname) fclose (fin);
    return i;
}

/* Read from a alternate libfile into the buffer and decompress the data
unsigned CPacker::ReadLibFile (const char *fname, unsigned num) {
    unsigned i;
    char     c;
    if (fname == NULL) return Str_ErrOpenR;
    fin = fopen (fname, "rb");
    if (fin == NULL) return Str_ErrOpenR;
    for (i=0; AppName[i] != 0; i++) {
        c = fgetc (fin);
        if (c != AppName[i]) {
            fclose (fin); return Str_Checksum;
        }
    }
    fseek (fin, -4 * num - 4, SEEK_END);
    fread (&i, 4, 1, fin);
    fseek (fin, i, SEEK_SET);
    i = ReadBufCore ();
    fclose (fin);
    return i;
}
*/

unsigned CPacker::ReadBuffer (FILE *file) {
    if (file == NULL) return Str_ErrOpenR;
    fin = file;
    return ReadBufCore ();
}

unsigned CPacker::ReadBuffer (void) {
    fin = theApp.MediaLib;
    if (fin == NULL) return Str_ErrOpenR;
    return ReadBufCore ();
}

unsigned CPacker::ReadBufferPure (const char *fname) {
    fin = fopen (fname, "rb");
    if (fin == NULL) return Str_ErrOpenR;
    fseek (fin, 0, SEEK_END);
    textsize = ftell (fin);
    if (textsize > MAXLIBSIZE) textsize = MAXLIBSIZE;
    rewind (fin);
    fread (Buffer, textsize, 1, fin);
    if (ferror (fin)) {
        textsize = 0;
        fclose (fin);
        return Str_ErrRead;
    }
    ramptr = Buffer;
    fclose (fin);
    return 0;
}

void CPacker::BufferCopy (unsigned char *dest) {
    if (dest) memcpy (dest, Buffer, textsize);
}



/* Read and write various data types to/from the buffer */
#ifndef MAC

static unsigned short inline SwapW (unsigned short n) { return n; }
static unsigned   inline SwapD (unsigned   n) { return n; }

#else

static unsigned short inline SwapW (unsigned short n) {
    return ((n & 0xff00) >> 8) | ((n & 0x00ff) << 8);
}

static unsigned inline SwapD (unsigned n) {
    return ((n & 0xff000000) >> 24) | ((n & 0x00ff0000) >> 8) | ((n & 0x0000ff00) << 8) | ((n & 0x000000ff) << 24);
}

#endif


void CPacker::WriteB (unsigned char d) {
    ramptr [0] = d;
    ramptr++;
    if (ramptr > Buffer + textsize) textsize = ramptr - Buffer;
}

void CPacker::WriteW (unsigned short d) {
    d = SwapW (d);
    ((short*) (ramptr)) [0] = d;
    ramptr += 2;
    if (ramptr > Buffer + textsize) textsize = ramptr - Buffer;
}

void CPacker::WriteD (unsigned d) {
    d = SwapD (d);
    ((int*) (ramptr)) [0] = d;
    ramptr += 4;
    if (ramptr > Buffer + textsize) textsize = ramptr - Buffer;
}

unsigned char CPacker::ReadB (void) {
    char d = ramptr  [0];
    ramptr += 1;
    return d;
}

unsigned short CPacker::ReadW (void) {
    short d = ((short*) ramptr) [0];
    ramptr += 2;
    return SwapW (d);
}

unsigned CPacker::ReadD (void) {
    int d = ((int*) ramptr) [0];
    ramptr += 4;
    return SwapD (d);
}

void CPacker::WriteDmulti (void *d, unsigned ln) {
    /* does not work! */
    unsigned i;
    ((char*) 0)[1] = 1;
    for (i=0; i<ln; i+= sizeof (unsigned)) WriteD (((unsigned*) d)[i]);
}

void CPacker::ReadDmulti (void *d, unsigned ln) {
    /* does not work! */
    unsigned i;
    ((char*) 0)[1] = 1;
    for (i=0; i<ln; i+= sizeof (unsigned)) ((unsigned*) d) [i] = ReadD ();
}

/* Remove for mac version! */
void CPacker::Write (void *d, unsigned ln) {
    if (d) memcpy (ramptr, d, ln);
    ramptr += ln;
    if (ramptr > Buffer + textsize) textsize = ramptr - Buffer;
}

void CPacker::Read (void *d, unsigned ln) {
    memcpy (d, ramptr, ln);
    ramptr += ln;
}


/* !Here starting all the internal routines don't call them from inside your app! */

/* Reads a byte from File/Ram/Spec.  (used in the ENcoder) */
int CPacker::SpecRead (void) {
    int ret = -1;
    if (pos < textsize) ret = ramptr [pos];
    pos++;
    return ret;
}

void CPacker::UpdateModel (int sym) {
    int i;
    
    if (sym_cum[0] >= MAX_CUM) {
        int c = 0;
        for (i=N_CHAR; i>0; i--) {
            sym_cum[i] = c;
            c+=(sym_freq[i] = (sym_freq[i] + 1) >> 1);
        }
        sym_cum[0] = c;
    }

    for (i = sym; sym_freq[i] == sym_freq[i - 1]; i--);
    if (i < sym) {
        int ch_i = sym_to_char[i];    /* swap */
        int ch_sym = sym_to_char[sym];
        sym_to_char[i] = ch_sym;
        sym_to_char[sym] = ch_i;
        char_to_sym[ch_i] = sym;
        char_to_sym[ch_sym] = i;
    }
    sym_freq[i]++;
    while (--i >= 0) sym_cum[i]++;
}

void CPacker::Decode (void) {
    int i, j, l, r, c;
    unsigned k;
    DWORD count;

    low      = 0;
    high     = Q4;
    value    = 0;
    shifts   = 0;
    pos      = 0;
    buffer   = 0;
    mask     = 128;
    codesize = sizeof (textsize);

    for (i=0; i<M+2; i++) {
        GetBit (buffer);
    }

    memcpy (char_to_sym, stdtable, (N_CHAR * 4) + N + 4);

    for (i = 0; i < N - F; i++) text_buf[i] = ' ';
    r = N - F;

    for (count = 0; count < textsize; ) {
        i = high - low;
        k = (WORD)(((value - low + 1) * sym_cum[0] - 1) / i);
        j = 1;
        l = N_CHAR;    
        while (j<l) {
            c = (j+l) / 2;
            if (sym_cum[c] > k) j = c + 1;
            else l = c;
        }
        high = low + (i * sym_cum[j - 1]) / sym_cum[0];
        low  +=      (i * sym_cum[j    ]) / sym_cum[0];
        for (;;) {
            if (low >= Q2) {
                value -= Q2;
                low   -= Q2;
                high  -= Q2;
            } else if (low >= Q1 && high <= Q3) {
                value -= Q1;
                low   -= Q1;
                high  -= Q1;
            } else if (high > Q2) break;
            low  += low;
            high += high;

            GetBit (buffer);
        }
        c = sym_to_char[j];
    
        if (sym_cum[0] >= MAX_CUM) {
            k = 0;
            for (i=N_CHAR; i>0; i--) {
                sym_cum[i] = k;
                k += (sym_freq[i] = (sym_freq[i] + 1) >> 1);
            }
            sym_cum[0] = k;
        }

        for (i = j; sym_freq[i] == sym_freq[i - 1]; i--);
        if (i < j) {
            l = sym_to_char[i];
            k = sym_to_char[j];
            sym_to_char[i] = k;
            sym_to_char[j] = l;
            char_to_sym[l] = j;
            char_to_sym[k] = i;
        }
        sym_freq[i]++;
        while (--i >= 0) sym_cum[i]++;

        if (c < 256) {
            ramptr   [pos] = c;
            pos++;
            text_buf [r++] = c;
            r &= (N - 1);
            count++;
        } else {
            k = (WORD)(((value - low + 1) * position_cum[0] - 1) / (high - low));
            i = 1;
            l = N;
            while (i < l) {
                j = (i + l) / 2;
                if (position_cum[j] > k) i = j + 1;
                else l = j;
            }
            j    = i - 1;
            k    = high - low;
            high = low + (k * position_cum[j    ]) / position_cum[0];
            low +=       (k * position_cum[j + 1]) / position_cum[0];
            for  (;;) {
                if (low >= Q2) {
                    value -= Q2;
                    low   -= Q2;
                    high  -= Q2;
                } else if (low >= Q1 && high <= Q3) {
                    value -= Q1;
                    low   -= Q1;
                    high  -= Q1;
                } else if (high > Q2) break;
                low  += low;
                high += high;
                GetBit (buffer);
            }
            i = (r - j - 1) & (N - 1);
            j = c - 255 + THRESHOLD;

            for (l = 0; l < j; l++) {
                c = text_buf[(i + l) & (N - 1)];
                ramptr   [pos] = c;
                pos++;
                text_buf [r++] = c;
                r &= (N - 1);
                count++;
            }
        }
    }
}


/* Internal Compress Routines (don't call them in your apps) */
void CPacker::PutBit (int bit) {
    if (bit) buffer |= mask;
    if ((mask <<= 1) == 256) {
        putc     (buffer, fout);
        mask     = 1;
        buffer   = 0;
        codesize++;
    }
}

void CPacker::InsertNode (int r) {
    int  i, p, cmp, temp;
    BYTE *key;

    cmp = 1;
    key = &text_buf[r];
    p = N + 1 + key[0];
    rson[r] = lson[r] = NIL;
    match_length = 0;
    for  (;;) {
        if (cmp >= 0) {
            if (rson[p] != NIL) p = rson[p];
            else {
                rson[p] = r;
                dad[r] = p;
                return;
            }
        } else {
            if (lson[p] != NIL) p = lson[p];
            else {
                lson[p] = r;
                dad[r] = p;
                return;
            }
        }
        for (i = 1; i < F; i++)
            if ((cmp = key[i] - text_buf[p + i]) != 0)  break;
        if (i > THRESHOLD) {
            if (i > match_length) {
                match_position = (r - p) & (N - 1);
                if ((match_length = i) >= F) break;
            } else if (i == match_length) {
                if ((temp = (r - p) & (N - 1)) < match_position)
                    match_position = temp;
            }
        }
    }
    dad[r]  = dad[p];
    lson[r] = lson[p];
    rson[r] = rson[p];
    dad[lson[p]] = r;
    dad[rson[p]] = r;
    if (rson[dad[p]] == p) rson[dad[p]] = r;
    else                   lson[dad[p]] = r;
    dad[p] = NIL;
}

void CPacker::DeleteNode (int p) {
    int q;
    if (dad[p] == NIL) return;
    if (rson[p] == NIL) q = lson[p];
    else if (lson[p] == NIL) q = rson[p];
    else {
        q = lson[p];
        if (rson[q] != NIL) {
            do {  q = rson[q];  } while (rson[q] != NIL);
            rson[dad[q]] = lson[q];
            dad[lson[q]] = dad[q];
            lson[q] = lson[p];
            dad[lson[p]] = q;
        }
        rson[q] = rson[p];
        dad[rson[p]] = q;
    }
    dad[q] = dad[p];
    if (rson[dad[p]] == p) rson[dad[p]] = q;
    else                   lson[dad[p]] = q;
    dad[p] = NIL;
}

void CPacker::Output (int bit) {
    PutBit(bit);
    for  (; shifts > 0; shifts--) PutBit(! bit);
}

void CPacker::EncodeChar (int ch) {
    int  sym;
    DWORD range;

    sym = char_to_sym[ch];
    range = high - low;
    high = low + (range * sym_cum[sym - 1]) / sym_cum[0];
    low +=       (range * sym_cum[sym    ]) / sym_cum[0];
    for (;;) {
        if (high <= Q2) Output(0);
        else if (low >= Q2) {
            Output(1);
            low -= Q2;
            high -= Q2;
        } else if (low >= Q1 && high <= Q3) {
            shifts++;
            low -= Q1;
            high -= Q1;
        } else break;
        low += low;
        high += high;
    }
    UpdateModel(sym);
}

void CPacker::EncodePosition (int position) {
    DWORD range;

    range = high - low;
    high = low + (range * position_cum[position    ]) / position_cum[0];
    low +=       (range * position_cum[position + 1]) / position_cum[0];
    for (;;) {
        if (high <= Q2) Output(0);
        else if (low >= Q2) {
            Output(1);
            low-=Q2;
            high-=Q2;
        } else if (low >= Q1 && high <= Q3) {
            shifts++;
            low-=Q1;
            high-=Q1;
        } else break;
        low+=low;
        high+=high;
    }
}

void CPacker::Encode (void) {
    int  i, c, len, r, s, last_match_length;

    low      = 0;
    high     = Q4;
    value    = 0;
    shifts   = 0;
    codesize = sizeof (textsize);
    pos      = 0;
    filepos  = 0;
    buffer   = 0;
    mask     = 1;

    sym_cum[N_CHAR] = 0;
    for (int sym = N_CHAR; sym >= 1; sym--) {
        int ch = sym - 1;
        char_to_sym[ch] = sym;
        sym_to_char[sym] = ch;
        sym_freq[sym] = 1;
        sym_cum[sym - 1] = sym_cum[sym] + sym_freq[sym];
    }
    sym_freq[0] = 0;
    position_cum[N] = 0;

    for (i = N; i >= 1; i--)
        position_cum[i - 1] = position_cum[i] + 10000 / (i + 200);

    for (i=N+1; i<=N+256; i++) rson[i]=NIL;
    for (i=0; i<N; i++) dad[i]=NIL;

    s = 0;  r = N - F;
    for (i = s; i < r; i++) text_buf[i] = ' ';
    for (len = 0; len < F && (c = SpecRead()) != EOF; len++)
        text_buf[r + len] = c;

    for (i = 1; i <= F; i++) InsertNode(r - i);
    InsertNode(r);
    do {
        if (match_length > len) match_length = len;
        if (match_length <= THRESHOLD) {
            match_length = 1;
            EncodeChar(text_buf[r]);
        } else { 
            EncodeChar(255 - THRESHOLD + match_length);
            EncodePosition(match_position - 1);
        }
        last_match_length = match_length;
        for (i = 0; i < last_match_length && (c = SpecRead()) != EOF; i++) {
            DeleteNode(s);
            text_buf[s] = c;
            if (s < F - 1) text_buf[s + N] = c;
            s = (s + 1) & (N - 1);
            r = (r + 1) & (N - 1);
            InsertNode(r);
        }

        while (i++ < last_match_length) {
            DeleteNode(s);
            s = (s + 1) & (N - 1);
            r = (r + 1) & (N - 1);
            if (--len) InsertNode(r);
        }
    } while (len > 0);
    shifts++;
    if (low < Q1) Output(0); else Output(1);
    for (i = 0; i < 7; i++) PutBit (0);
}
